import threading
from collections import deque
import time
from DrissionPage import ChromiumPage, ChromiumOptions
from DrissionPage._pages.chromium_tab import ChromiumTab
from DrissionPage._units.listener import DataPacket
from typing import Tuple, Optional
import json


class DPProxyClient:
    def __init__(self, driver: ChromiumTab, self_kill=False):
        self.tab_id = driver.tab_id
        self.driver = ChromePageProxy(driver, self)
        self.thread = None
        self.self_kill = self_kill
        # self.packet_list = []
        self.packet_queue = deque()

    def get_driver(self, start_listen, count=None, timeout=10) -> ChromiumTab:
        """
        获取代理后的driver、tab
        :param start_listen: 若你自己写的代码里已经使用自动化框架监听发包的功能了，则该值应该置为False。若没监听，则必须将该值置为True
        :param count: 需捕获的数据包总数，为None表示无限
        :param timeout: 两个数据包之间等待的最大时长（秒），为None表示无限，默认为10秒
        :return:
        """
        if start_listen:
            self.driver.listen.set_targets(res_type="XHR")
            self.driver.listen.start()
            self.thread = threading.Thread(target=self.start_listen, args=(count, timeout,))
            self.thread.start()
        return self.driver

    def start_listen(self, count=None, timeout=10):
        for _ in self.driver.listen.steps(count=count, timeout=timeout, gap=1):
            pass

    # 每次调用函数，都从队列的左端弹出一个数据包
    def pop_first_packet(self):
        if self.packet_queue:
            result = self.packet_queue.popleft()
            return json.dumps(result, ensure_ascii=False)
        else:
            return None


class DPProxyClientManager:
    """浏览器池管理器 - 使用单例模式"""
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.tab_pool = {}
        return cls._instance

    def create_client(self, tab: ChromiumTab, self_kill=False) -> Tuple[str, DPProxyClient, ChromiumTab]:
        """创建新的tab页面代理实例"""
        client = DPProxyClient(tab, self_kill=self_kill)
        tab = client.get_driver(True)
        tab_id = tab.tab_id
        self.tab_pool[tab_id] = {"client": client, "driver": tab}
        return tab_id, client, tab

    def get_client(self, tab_id: str) -> Optional[DPProxyClient]:
        """根据端口获取浏览器实例"""
        return self.tab_pool.get(tab_id).get("client", None)

    def remove_client(self, tab_id: str) -> Tuple[bool, Optional[ChromiumPage]]:
        """根据端口移除浏览器实例"""
        client = self.tab_pool.pop(tab_id, None)
        return client is not None, client

    def list_clients(self) -> list[int]:
        """列出所有活跃的浏览器端口"""
        return list(self.tab_pool.keys())


class ChromePageProxy:
    def __init__(self, page, client=None):
        self.__dict__['page'] = page
        self.__dict__['client'] = client

    def __getattr__(self, item):
        attr = getattr(self.page, item)
        print(item, attr)
        if item == 'listen':
            listen_proxy = DrissionPageListenerProxy(attr, self.__dict__['client'])
            return listen_proxy
        return attr


class DrissionPageListenerProxy:
    def __init__(self, listener, client=None):
        self.listener = listener
        self.client = client

    def __getattr__(self, item):
        attr = getattr(self.listener, item)
        # 当监听到wait被调用的时候
        if item == "wait":
            def wrapper(*args, **kwargs):
                result = attr(*args, **kwargs)
                check_data_packet(result, self.client)
                return result

            return wrapper
        # 当监听到steps被调用的时候
        if item == "steps":
            def wrapper(*args, **kwargs):
                if kwargs.get("gap", 1) > 1:
                    raise Exception("暂不支持多包监控")
                result = attr(*args, **kwargs)
                if attr.__name__ == "steps":
                    for step in result:
                        check_data_packet(step, self.client)
                        yield step

            return wrapper
        return attr


def check_data_packet(packet: DataPacket, client: DPProxyClient):
    """
    封装监听到的数据包，并将其存放在client的packet_queue中
    :param packet:
    :param client:
    :return:
    """
    url = packet.url
    method = packet.request.method
    data = None
    if packet.request.hasPostData:
        data = packet.request.postData
    body = packet.response.body
    temp_dict = {
        "url": url,
        "method": method,
        "request_data": data,
        "request_headers": dict(packet.request.headers),
        "response_headers": dict(packet.response.headers),
        "response_body": body,
    }
    # print(temp_dict)
    client.packet_queue.append(temp_dict)


client_manager = DPProxyClientManager()

if __name__ == '__main__':
    co = ChromiumOptions().set_user_agent(
        "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/143.0.0.0 Mobile Safari/537.36")
    tab = ChromiumPage(co).latest_tab
    client = DPProxyClient(tab, self_kill=False)
    # client = CaptchaClient(tab, self_kill=True)
    tab = client.get_driver(True)
    url = "https://api.toutiaoapi.com/feoffline/hotspot_and_local/html/hot_list/index.html?client_extra_params=%7B%22custom_log_pb%22%3A%22%7B%5C%22style_id%5C%22%3A%5C%2240030%5C%22%2C%5C%22entrance_hotspot%5C%22%3A%5C%22search%5C%22%2C%5C%22location%5C%22%3A%5C%22hot_board%5C%22%2C%5C%22category_name%5C%22%3A%5C%22hotboard_light%5C%22%7D%22%7D&count=50&log_pb=%7B%22style_id%22%3A%2240030%22%2C%22entrance_hotspot%22%3A%22search%22%2C%22location%22%3A%22hot_board%22%2C%22category_name%22%3A%22hotboard_light%22%7D&only_hot_list=1&tab_name=stream&enter_keyword=%23%E7%BE%8E%E5%9B%BD%E9%80%80%E5%87%BA66%E4%B8%AA%E5%9B%BD%E9%99%85%E7%BB%84%E7%BB%87%23"
    tab.get(url)
    for _ in range(5056):
        new_packet = client.pop_first_packet()
        print(new_packet, "23")
        time.sleep(1)



import time
import random
import os
import minify_html
from DrissionPage import ChromiumPage, ChromiumOptions
from bs4 import BeautifulSoup
from curl_cffi import requests
from lxml import html, etree


# 使用requests获取html，用于测试是否使用了瑞数和jsl
def requests_html(url):
    headers = {
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/142.0.0.0 Safari/537.36",
    }
    response = requests.get(url, headers=headers, verify=False)
    response.encoding = "utf-8"
    return response.text, response.status_code


# 使用dp无头模式获取html，用于测试是否使用了其他waf，如移动waf
def dp_headless_html(url):
    opt = ChromiumOptions().headless(True)
    opt.set_argument('--no-sandbox')
    """创建新的浏览器实例"""
    random_port = random.randint(9934, 10034)
    custom_data_dir = os.path.join(os.path.expanduser('~'), 'DrissionPage', "userData", f"{random_port}")
    opt.set_user_data_path(custom_data_dir)  # 设置用户数据路径
    opt.set_local_port(random_port)
    page = ChromiumPage(opt)
    tab = page.latest_tab
    tab.get(url)
    # todo: 目前没有更好的方式，为了数据渲染完全，只能硬等【受网速波动影响比较大】
    time.sleep(10)
    page_html = tab.html
    # 无头浏览器在用完之后一定要记得再page级别进行quit
    page.quit()
    return page_html


# 压缩html
def compress_html(content, only_text=False):
    doc = html.fromstring(content)
    # 删除 style 和 script 标签
    for element in doc.xpath('//style | //script'):
        element.getparent().remove(element)

    # 删除 link 标签
    for link in doc.xpath('//link[@rel="stylesheet"]'):
        link.getparent().remove(link)

    # 删除 meta 标签（新增功能）
    for meta in doc.xpath('//meta'):
        meta.getparent().remove(meta)

    # 删除 style 属性
    for element in doc.xpath('//*[@style]'):
        element.attrib.pop('style')

    # 删除所有 on* 事件属性
    for element in doc.xpath('//*'):
        for attr in list(element.attrib.keys()):
            if attr.startswith('on'):
                element.attrib.pop(attr)

    result = etree.tostring(doc, encoding='unicode')
    result = minify_html.minify(result)
    compress_rate = round(len(content) / len(result) * 100)
    print(f"html压缩比=> {compress_rate}%")
    if not only_text:
        return result, compress_rate
    soup = BeautifulSoup(result, 'html.parser')
    result = soup.get_text(strip=True)
    return result, compress_rate


# 通过cookie判断是否有waf，需要通过遇到的例子，不断的完善cookie判别函数
def assert_waf_cookie(cookies: list):
    for cookie in cookies:
        cookie_name = cookie['name']
        cookie_value = cookie['value']
        if len(cookie_name) == 13 and len(cookie_value) == 88:
            return True, "瑞数"
        if "_jsl" in cookie_name:
            return True, "加速乐"
    return False, "没有waf"


# 对dp_mcp的消息打包
def dp_mcp_message_pack(message: str, **kwargs):
    text_obj = {key: value for key, value in kwargs.items()}
    text_obj.update({"message": message})
    return {
        "content": [{
            "type": "text",
            # "text": json.dumps(text_obj, ensure_ascii=False)
            "text": text_obj
        }]
    }

# todo: 大致盘一下各种判定的逻辑【以下的所有压缩比之间的差距均取“绝对值”】
#  1. 如果requests、无头、有头获取到的压缩比之间从差距都在15%以内，则认定该页面是静态页面，此时优先使用requests请求
#  2. 如果requests的status_code为特定的412，或者521，则判定是瑞数和jsl。[此时还有一个特点：requests的压缩比会与其他两种方式获取到的压缩比差距非常大(一两千的那种)]
#  3. 如果requests、无头、有头获取到的压缩比之间差距都在40%以上，则判定该页面只可以用有头采集
#  4. 如果无头和有头获取到的压缩比之间差距小于15%，但是requests和无头的差距大于40%，则认定该页面可以使用无头浏览器采集
#  5. 如果requests和有头获取到的压缩比之间差距小于15%，但是无头和有头的差距大于40%，则认定该页面优先使用有头浏览器采集
#  【此时可能是：1.使用了别的检测无头的waf。2.网站使用瑞数，但是这次请求没有拦截requests（不知道是不是瑞数那边故意设置的），
#   此时如果想进一步判定是否是瑞数，可以使用有头浏览器取一下cookies，如果cookies里面存在瑞数的cookie，那么就可以断定是瑞数】



import hashlib
import json
import os
from typing import Any

from fastmcp import FastMCP

from tools.browser_manager import BrowserManager
from tools.tools import compress_html, requests_html, dp_headless_html, assert_waf_cookie, dp_mcp_message_pack
from tools.browser_proxy import DPProxyClient, DPProxyClientManager

html_source_code_local_save_path = os.path.join(os.getcwd(), "html-source-code")
waf_status_code_dict = {
    412: "瑞数",
    521: "加速乐"
}
# 一轮最大输入，以免单个html最大长度超过ai最大输入
ont_turn_max_token = 25000


def register_visit_url(mcp: FastMCP, browser_manager: BrowserManager, client_manager: DPProxyClientManager):
    @mcp.tool(name="visit_url",
              description="使用Drissionpage打开url访问某个网站，并开始监听初始tab页的所有的XHR请求，当需要使用手机版浏览器Ua时use_mobile_user_agent为True")
    async def visit_url(url: str, use_mobile_user_agent=False) -> dict[str, Any]:
        mobile_user_agent = None
        if use_mobile_user_agent:
            mobile_user_agent = "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/143.0.0.0 Mobile Safari/537.36"
        port, _browser = browser_manager.create_browser(mobile_user_agent)
        tab = _browser.get_tab()
        client_manager.create_client(tab)
        tab.get(url)
        tab_id = tab.tab_id
        return dp_mcp_message_pack(
            f"已在[{port}]端口创建浏览器对象，并已打开链接：{url}，打开的模式是：{'手机版' if use_mobile_user_agent else '电脑版'}",
            tab_id=tab_id,
            browser_port=port
        )


def register_get_html(mcp: FastMCP, browser_manager, client_manager: DPProxyClientManager):
    @mcp.tool(name="get_html", description="使用Drissionpage获取某一个tab页的html")
    async def get_html(browser_port: int, tab_id: str) -> dict[str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        tab = _browser.get_tab(tab_id)
        file_name_prefix = hashlib.md5(str(tab.title).encode('utf-8')).hexdigest()
        if not os.path.exists(html_source_code_local_save_path):
            os.makedirs(html_source_code_local_save_path)
        min_html, compress_rate = compress_html(tab.html)
        html_str_list = [min_html[i:i + ont_turn_max_token] for i in range(0, len(min_html), ont_turn_max_token)]
        html_file_list = []
        for index, html_str in enumerate(html_str_list):
            file_name = file_name_prefix + f"_{tab_id}_segment{index}.html"
            abs_path = os.path.join(html_source_code_local_save_path, file_name)
            with open(abs_path, "w", encoding="utf-8") as f:
                f.write(html_str)
            html_file_list.append(abs_path)
        message = f"已保存tab页：【{tab_id}】的html源码片段共{len(html_file_list)}个"
        return dp_mcp_message_pack(message, tab_id=tab_id, htmls_local_path=html_file_list)


def register_get_new_tab(mcp: FastMCP, browser_manager, client_manager: DPProxyClientManager):
    @mcp.tool(name="get_new_tab",
              description="使用Drissionpage创建一个新的tab页，在新的tab页中打开url，并开始监听新的tab页的所有XHR请求")
    async def get_new_tab(browser_port: int, url: str) -> dict[str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        tab = _browser.new_tab()
        client_manager.create_client(tab)
        tab.get(url)
        _browser.activate_tab(tab)
        tab_id = tab.tab_id
        return dp_mcp_message_pack(f"已创建新的tab页，并打开链接：{url}", tab_id=tab_id)


def register_switch_tab(mcp: FastMCP, browser_manager, client_manager: DPProxyClientManager):
    @mcp.tool(name="switch_tab", description="根据传入的tab_id切换到对应的tab页", )
    async def switch_tab(browser_port: int, tab_id: str) -> dict[str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        _browser.activate_tab(tab_id)
        return dp_mcp_message_pack(f"已将tab页:【{tab_id}】切换至最前端")


def register_close_tab(mcp: FastMCP, browser_manager, client_manager: DPProxyClientManager):
    @mcp.tool(name="close_tab", description="根据传入的tab_id关闭tab页", )
    async def close_tab(browser_port, tab_id) -> dict[str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        _browser.close_tabs(tab_id)
        return dp_mcp_message_pack(f"已将tab页:【{tab_id}】关闭")


def register_check_selector(mcp: FastMCP, browser_manager, client_manager: DPProxyClientManager):
    @mcp.tool(name="check_selector", description="查找tab页中是否包含元素，并返回元素attr_name所对应的值")
    async def check_selector(browser_port: int, tab_id: str, css_selector: str, attr_name: str = "text") -> dict[
        str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        target_tab = _browser.get_tab(tab_id)
        css_selector = css_selector
        if "css:" not in css_selector:
            css_selector = "css:" + css_selector
        target_eles = target_tab.eles(css_selector)
        exist_flag = False
        if len(target_eles) != 0:
            exist_flag = True
        if attr_name == "text":
            ele_text_list = [i.text.replace("\n", "") for i in target_eles]
            attr_output = "\n".join(ele_text_list)
        else:
            attr_output = json.dumps([i.attr(attr_name) for i in target_eles])
        return dp_mcp_message_pack(
            f"已完成tab页:【{tab_id}】对：【{css_selector}】的检查",
            tab_id=tab_id,
            selector=css_selector,
            selector_ele_exist=exist_flag,
            attr_output=attr_output
        )


def register_quit_browser(mcp: FastMCP, browser_manager, client_manager: DPProxyClientManager):
    @mcp.tool(name="quit_browser", description="退出浏览器会话，关闭浏览器")
    async def quit_browser(browser_port: int) -> dict[str, Any]:
        flag, _browser = browser_manager.remove_page(browser_port)
        if flag:
            _browser.quit()
        return dp_mcp_message_pack(
            f"浏览器[{browser_port}]，退出会话，关闭浏览器{'成功' if flag else '失败'}",
            browser_port=browser_port,
            quit_flag=flag
        )


def register_pop_first_packet(mcp: FastMCP, browser_manager, client_manager: DPProxyClientManager):
    @mcp.tool(name="pop_first_packet",
              description="每调用一次就会弹出传入的tab页所监听到的数据包中的第一个packet_message")
    async def pop_first_packet(browser_port: int, tab_id: str) -> dict[str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        client = client_manager.get_client(tab_id)
        packet_message = client.pop_first_packet()
        message = f"tab页:【{tab_id}】，暂时没有监听到XHR数据包"
        if packet_message:
            message = f"tab页:【{tab_id}】，监听到XHR数据包",
        return dp_mcp_message_pack(
            message,
            browser_port=browser_port,
            tab_id=tab_id,
            packet_message=packet_message
        )


def register_assert_waf(mcp: FastMCP, browser_manager, client_manager: DPProxyClientManager):
    @mcp.tool(name="assert_waf",
              description="通过对比requests、有头浏览器、无头浏览器获取到的html，判断网页是否使用了waf以及是否为动态渲染的网页")
    async def assert_waf(browser_port: int, tab_id: str) -> dict[str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        target_tab = _browser.get_tab(tab_id)
        recommend_team = "drissionpage_head"
        head_cookies = target_tab.cookies()
        # 通过cookie判断是否有waf
        waf_flag, waf_type = assert_waf_cookie(head_cookies)
        head_html = target_tab.html
        min_head_html, head_rate = compress_html(head_html, only_text=True)
        raw_html, status_code = requests_html(target_tab.url)
        min_raw_html, raw_rate = compress_html(raw_html, only_text=True)
        r_h_rate_diff = abs(head_rate - raw_rate)
        # 如果有已知的防火墙，则不浪费时间使用无头获取html和压缩比了
        if waf_flag or status_code in waf_status_code_dict.keys():
            return dp_mcp_message_pack(
                f"已完成tab页:【{tab_id}】的分析，该tab页存在waf",
                tab_id=tab_id,
                recommend_team=recommend_team,
                raw_head_rate_difference=r_h_rate_diff,
                raw_headless_rate_difference=0,
                head_headless_rate_difference=0
            )

        headless_html = dp_headless_html(target_tab.url)
        min_headless_html, headless_rate = compress_html(headless_html, only_text=True)
        r_hless_rate_diff = abs(raw_rate - headless_rate)
        h_hless_rate_diff = abs(head_rate - headless_rate)
        # 最优情况：requests，dp有头和无头拿到的结果基本一致，认定为没有防护的静态网页
        if r_h_rate_diff < 40 and r_hless_rate_diff < 40 and h_hless_rate_diff < 40:
            recommend_team = "requests"
        # 最差情况：requests，dp有头和无头拿到的结果差距都很大，认定为有浏览器无头检测+动态网页
        # if r_h_rate_diff < 40 and r_hless_rate_diff < 40 and h_hless_rate_diff < 40:
        # 较差1：dp有头和无头差距很小，但是requests拿不到正确结果，认定为有requests防护 or 动态网页
        elif h_hless_rate_diff < 30 and r_hless_rate_diff > 40:
            recommend_team = "drissionpage_headless"
        # 较差2：有头和无头差距很大，但是requests和有头拿到的结果基本一致，认定为要么有别的没有防护requests的waf，或者间歇性的瑞数【此时应该拿有头的cookie去判断其中是否有瑞数特征，上面已经做了】
        # if r_h_rate_diff < 15 and h_hless_rate_diff > 40:
        return dp_mcp_message_pack(
            f"已完成tab页:【{tab_id}】的分析，该tab页存在waf",
            tab_id=tab_id,
            recommend_team=recommend_team,
            raw_head_rate_difference=r_h_rate_diff,
            raw_headless_rate_difference=h_hless_rate_diff,
            head_headless_rate_difference=h_hless_rate_diff
        )
